/***********************************************************************

  Copyright (C) 2006  Adrien Brochier <a.brochier@free.fr>

  This file is part of SADE (Sivia Applied to Dielectric spectroscopy Experiment).

  SADE is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published
  by the Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.

  SADE is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA  02111-1307  USA

************************************************************************/

/**
  \file debye.h
  \author Adrien Brochier <a.brochier@free.fr>
  \date 07-06
 */

#include "pave.h"

#define EVAL 0
#define CONTRACT 1

const interval EMPTY=interval(1,-1);
const REEL permi=8.854187e-12;
/**
\brief evaluation of a debye sum.

\param x is a constant real number, the frequence.
\param param is a pave of parameter
\param nbRelaxation is the number of relaxation
\return return a pave of size 2 : {real part, imaginary part} who is the result of the debye sum
*/

inline pave eval(const REEL& x, pave& param, const int& nbRelaxation)
{
	if(isEmpty(param)) return param;
	int i;
	pave epsilon;
	epsilon.resize(2);
	interval tempo;
	epsilon[0]=param[0];
	epsilon[1]=interval(0,0);
	for(i=0;i<nbRelaxation;++i)
	{
		tempo=((param[2*i+1]*DEUX_PI*x)^2)+1;
		epsilon[0]=epsilon[0]+(param[2*(1+i)]/(tempo));
		epsilon[1]=epsilon[1]+((param[2*(1+i)]*param[2*i+1]*DEUX_PI*x)/(tempo));
		
			
	}
	epsilon[1]=epsilon[1]+(param[2*nbRelaxation+1]/(permi*DEUX_PI*x));
	
	return epsilon;
}
	
/**
\brief reduce the size of param with respect to some constraints

\param x is a constant real number, the frequence.
\param param is a pave of parameter
\param nbRelaxation is the number of relaxation
\param yr is a real number who is the constraint for the real part of epsilon
\param yi is a real number who is the constraint for the imaginary part of epsilon
\return return the contracted pave of parameter
*/
inline pave contract(const REEL& x, const pave& param,const interval& yr,const interval& yi, const int& nbRelaxation)
{
	pave result;
	
	if(isEmpty(param)) 
	{
		result.push_back(EMPTY);
		return result;
	}
	
	//init and variable declarations. every variable have to appear only one time. 
	vector<interval> epsilon;
	vector<vector<interval> > t;
	vector<vector<interval> > z;
	vector<vector<interval> > partial;
	
	interval epsPrime(0,0);
	interval epsSecond(0,0);
	interval withoutConduct(0,0);
	interval temp(0,0);
	interval sigma=param[2*nbRelaxation+1];
	
	int i;
	epsilon.resize(nbRelaxation+1);
	t.resize(nbRelaxation);
	z.resize(nbRelaxation);
	partial.resize(nbRelaxation);
	


/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////FORWARD PROPAGATION/////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

	epsilon[0]=param[0];
	for(i=0;i<nbRelaxation;++i)
	{
		
		epsilon[i+1]=param[2*(1+i)];
		
		t[i].push_back(param[1+2*i]);
		t[i].push_back(param[1+2*i]);
		
		z[i].resize(4);
		partial[i].resize(2);
	
		///// z[4] //////
		z[i][0]=t[i][0]*DEUX_PI*x;
		
		z[i][1]=z[i][0]^2;
		
		z[i][2]=z[i][1]+1;
		
		z[i][3]=epsilon[i+1]/z[i][2];
		
		if(i==0)
		{
			//// eps' ////
			partial[i][0]=z[i][3];
			
			//// eps'' ////
			partial[i][1]=z[i][3]*t[i][1]*x*DEUX_PI;
		}else{
			//// eps' ////
			partial[i][0]=partial[i-1][0]+z[i][3];
			
			//// eps'' ////
			partial[i][1]=partial[i-1][1]+z[i][3]*t[i][1]*x*DEUX_PI;
		}
			
	}
	
	epsPrime=partial[nbRelaxation-1][0]+epsilon[0];
	
	withoutConduct=partial[nbRelaxation-1][1];
	epsSecond=withoutConduct+(sigma/(permi*DEUX_PI*x));

	

	
	
	
	

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////INCLUSION TEST//////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
	if((epsPrime < yr) && (epsSecond < yi) )
	{
		
		return param;
	}
	epsPrime=epsPrime&yr;
	epsSecond=epsSecond&yi;
	

	if((epsPrime.isEmpty())||(epsSecond.isEmpty()))
	{
		result.push_back(EMPTY); return result;
	}
	/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////BACKWARD PROPAGATION/////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
	partial[nbRelaxation-1][0]=(epsPrime-epsilon[0])&partial[nbRelaxation-1][0];
	epsilon[0]=(epsPrime-partial[nbRelaxation-1][0])&epsilon[0];

	sigma=((epsSecond-withoutConduct)*(permi*DEUX_PI*x))&sigma;
	withoutConduct=(epsSecond-(sigma/(permi*DEUX_PI*x)))&withoutConduct;
	
	partial[nbRelaxation-1][1]=withoutConduct&(partial[nbRelaxation-1][1]);
	
	for(i=nbRelaxation-1;i>=0;--i)
	{
		
		
		if(i==0)
		{
			z[i][3]=partial[i][0]&z[i][3];
			
			t[i][1]=(partial[i][1]/(z[i][3]*x*DEUX_PI))&t[i][1];			
			z[i][3]=(partial[i][1]/(t[i][1]*x*DEUX_PI))&z[i][3];
			
		}else{
			
			
			partial[i-1][0]=(partial[i][0]-z[i][3])&partial[i-1][0];
			z[i][3]=(partial[i][0]-partial[i-1][0])&z[i][3];
			
			partial[i-1][1]=(partial[i][1]-z[i][3]*t[i][1]*x*DEUX_PI)&partial[i-1][1];
			z[i][3]=((partial[i][1]-partial[i-1][1])/(t[i][1]*x*DEUX_PI))&z[i][3];
			t[i][1]=((partial[i][1]-partial[i-1][1])/(z[i][3]*x*DEUX_PI))&t[i][1];
		}
		
	
	
		epsilon[i+1]=(z[i][2]*z[i][3])&epsilon[i+1];
		
		z[i][2]=(epsilon[i+1]/z[i][3])&z[i][2];
		
		
		
		
		z[i][1]=(z[i][2]+(-1))&z[i][1];
		z[i][0]=(z[i][1]^(0.5))&z[i][0];
		
		
		t[i][0]=(z[i][0]/(DEUX_PI*x))&t[i][0];
		/*cout <<"___________" << endl;
		affiche(z[i][4]);
		affiche(partial[i][0]);
		affiche(partial[i][1]);
		affiche(z[i][0]);
		affiche(t[i][0]);
		//*/
		
	}


	//equalities constraints
	
	for(i=0;i<nbRelaxation;++i)
	{
		t[i][0]=t[i][0]&t[i][1];
		
	}
	for(i=0;i<nbRelaxation-1;++i)
	{
		t[i+1][0]=interval(max(t[i+1][0].getFirst(),t[i][0].getFirst()),t[i+1][0].getLast());
		
		result.push_back(epsilon[i]);
		result.push_back(t[i][0]);
	}
	
	result.push_back(epsilon[nbRelaxation-1]);
	result.push_back(t[nbRelaxation-1][0]);
	result.push_back(epsilon[nbRelaxation]);
	result.push_back(sigma);

	//return new parameter.

	return result;

	
};
	
