/***********************************************************************

  Copyright (C) 2006  Adrien Brochier <a.brochier@free.fr>

  This file is part of SADE (Sivia Applied to Dielectric spectroscopy Experiment).

  SADE is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published
  by the Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.

  SADE is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA  02111-1307  USA

************************************************************************/

/**
  \file pave.h
  \author Adrien Brochier <a.brochier@free.fr>
  \date 07-06
 */

#ifndef INTERVAL_H
#define INTERVAL_H
#include <vector>
#include <iostream>
#include <math.h>
#include <cstdlib>
#include <limits>
#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b))

/**
\typedef REEL
type used for Real numbers
*/
typedef double REEL;

/*some usefull constants*/
const REEL error=std::numeric_limits<REEL>::epsilon();/**< smallest real which verify 	1+epsilon != 1*/
const REEL PI2=3.1415926535897/2;
const REEL TROIS_PI2 = 3*PI2;
const REEL CINQ_PI2 = 5*PI2;
const REEL SEPT_PI2 = 7*PI2;
const REEL DEUX_PI = 4*PI2;
const REEL oo=1e10;
using namespace std;

/**
\class interval
\brief interval basic implementation

This class implement the concept of interval : we can use it as we use real number (+,/,-,*). there is also some ensemblist operations (union, intersection)..
*/
class interval
{
	/* friend functions =================================================== */
	
	/*elementary operation*/
	friend interval operator/(REEL a, interval& i);
	friend interval operator/(interval i, REEL a);
	friend interval operator-(REEL a, interval i);
	
	/*trigonometry*/
	friend interval iSinInv(interval i);
	friend interval iCosInv(interval i);
	friend interval iLog(interval i);
	friend interval iTan(interval i);

	/* Private variable =================================================== */
	
protected:
	/**lower bound*/
	REEL first;
	/**upper bound*/
	REEL last; 


	/* public functions =================================================== */

public:
	/**basic constructor. (*this) will be an ampty interval*/
	interval();
	/**construct the interval [f,l]*/
	interval(REEL f, REEL l);

	/*ensemblist operation*/
	/**intersection*/
	interval operator&(interval i);
	/**convex union : return the smallest interval containing (*this) Union (i)*/
	interval operator|(interval i);

	/*elementary operations*/
	interval operator+(interval i);
	interval operator+(REEL a);
	interval operator*(interval i);
	interval operator*(REEL a);
	interval operator-(interval i);
	interval operator/(interval i);
	interval operator^(REEL a);
	interval operator^(interval i);	
	interval operator%(REEL a);		/**<modulo*/
	int operator<(interval i);		/**<inclusion test*/
	void operator=(interval i);		/**<affectation*/

	/*class data access*/
	REEL getFirst();
	REEL getLast();
	REEL length();
	friend ostream& operator<<(ostream& o, interval i);		/**<output stream*/
	int isEmpty();					/**<test if (*this) is empty. \return 1 if (*this) is empty, 0 otherwise*/
};

/**
\typedef pave
\brief vector of interval
we define a "pave" as a vector of interval.
*/
typedef vector<interval> pave;



/* some public functions =================================================== */


/*inline void affiche(interval p)
{
	std::cout <<"[" << p.getFirst() << "," << p.getLast() << "]"<< endl;
};*/
/**cut 'in' in 2 part and place it in 'out1' and 'out2'*/
inline int bissection(pave in, pave& out1, pave& out2)
{
	
	int n=in.size()-1;
	int i,j;
	REEL length=-1;
	REEL l;
	
	for(i=0;i<n;i++)		//searching the index of the greatest interval
	{
		if(i%2==0)
		{
			l=in[i].length();		//it's an epsilon -> 'normal' cut
		}else{
			l=(log10(in[i].getLast())-log10(in[i].getFirst()));	//it's a tau -> 'logarithmic' cut
			
		}
		if(l > length )
		{
			j=i;
			length=l;
		}
	}
	
	
	out1=out2=in;
	REEL middle;
	if(length < in[n].length())		//case of the conductivity
	{
		middle=in[n].length()/2;
		out1[n]=interval(out1[n].getFirst(),out1[n].getFirst()+middle);
		out2[n]=interval(out2[n].getFirst()+middle,out2[n].getLast());	//cutting the interval in 2 equal part.
		
	} else{ //devrait y avoir un else, ici ??? non ???

	
		if(j%2==0)
		{
			middle=length/2;
			out1[j]=interval(out1[j].getFirst(),out1[j].getFirst()+middle);
			out2[j]=interval(out2[j].getFirst()+middle,out2[j].getLast());		//cutting the interval in 2 equal part.
		}else{
			middle=pow(10,length/2);
			out1[j]=interval(out1[j].getFirst(),out1[j].getFirst()*middle);
			out2[j]=interval(out2[j].getFirst()*middle,out2[j].getLast());		//cutting the interval in 2 "logarithmically equal" part.
		
		}
	}
		
	return 0;
	
};
inline int bissectionTau(pave in, pave& out1, pave& out2)
{
	
	int n=in.size()-1;
	int i,j;
	REEL length=-1;
	REEL l;
	
	for(i=1;i<n;i+=2)		//searching the index of the greatest interval
	{
		
		l=(log10(in[i].getLast())-log10(in[i].getFirst()));	//it's a tau -> 'logarithmic' cut
			
		if(l > length )
		{
			j=i;
			length=l;
		}
	}
	
	
	out1=out2=in;
	REEL middle;
	
	
	if(j%2==0)
	{
		middle=length/2;
		out1[j]=interval(out1[j].getFirst(),out1[j].getFirst()+middle);
		out2[j]=interval(out2[j].getFirst()+middle,out2[j].getLast());		//cutting the interval in 2 equal part.
	}else{
		middle=pow(10,length/2);
		out1[j]=interval(out1[j].getFirst(),out1[j].getFirst()*middle);
		out2[j]=interval(out2[j].getFirst()*middle,out2[j].getLast());		//cutting the interval in 2 "logarithmically equal" part.
		
	}
		
	return 0;
	
};

inline REEL length(pave p)/**<return the length of a pave, ie the length og the greatest interval in the pave*/
{
	int i;
	REEL m=-1;
	REEL l;
	
	for (i=0;i<p.size()-1;i++)
	{
		if(i%2==0)
		{
			l=p[i].length();
		}else{
			l=(log10(p[i].getLast())-log10(p[i].getFirst()));
			/*if(l<0)
			{
				cout << l << endl;
				affiche(p[i]);
			}*/
			//l=p[i].length();
		}
		m=max(m,l);
	}
	m=max(m,p[p.size()-1].length());
	return m;
};
inline REEL lengthTau(pave p)/**<return the length of a pave, ie the length og the greatest interval in the pave*/
{
	int i;
	REEL m=-1;
	REEL l;
	
	for (i=1;i<p.size()-1;i+=2)
	{
			l=(log10(p[i].getLast())-log10(p[i].getFirst()));
			/*if(l<0)
			{
				cout << l << endl;
				affiche(p[i]);
			}*/
			//l=p[i].length();
		
		m=max(m,l);
	}
	return m;
};

/**test if p1 is ''smaller'' than p2*/
inline int mieux(pave p1,pave p2)
{
	
	
	REEL l1,l2;
	int i;
	
	for (i=0;i<p1.size()-1;i++)
	{
		if(i%2==0)
		{
			l1=p1[i].length();
			l2=p2[i].length();
		}else{
			l1=(log10(p1[i].getLast())-log10(p1[i].getFirst()));
			l2=(log10(p2[i].getLast())-log10(p2[i].getFirst()));
			
		}
		if(l1<=0) return 0;
		if(l1<=0.9*l2) return 1;		//here is the test
	}
	if(p1[p1.size()-1].length() <= 0.9*p2[p1.size()-1].length()) return 1;
	return 0;
};



/**convex union of 2 'pave'*/
inline pave operator|(pave p1, pave p2)
{
	
	
	pave r;
	int i;
	if(p1.empty()) return p2;
	if(p2.empty()) return p1;
	
	for (i=0;i<p1.size();i++)
	{
		r.push_back(p1[i]|p2[i]);
	}
	return r;
};
inline pave operator&(pave p1, pave p2)/**<intersection of 2 pave*/
{
	
	
	pave r;
	int i;
	
	for (i=0;i<p1.size();i++)
	{
		r.push_back(p1[i]&p2[i]);
	}
	return r;
};



inline int isEmpty(pave p)/**<test if p is empty*/ 	
{
	
	
	if(p.empty()) return 1;
	for (int i=0; i <p.size();i++)
	{
		if(p[i].isEmpty())
		{
			return 1;
		}
	}
	return 0;
};

/**print a pave to the standard output*/
inline void affiche(pave p)
{	if(!isEmpty(p))
	{
		unsigned int i;
		std::cout <<"[" << p[0].getFirst() << "," << p[0].getLast() << "]";
		for (i=1;i<p.size();i++)
		{
			std::cout <<"x" <<"[" << p[i].getFirst() << "," << p[i].getLast() << "]";
		}
		std::cout << std::endl;
	}else{
		cout << "vide" << endl;
	}
};

inline int operator<(pave p1, pave p2)/**<test if p1 is included in p2*/
{
	
	
	if(isEmpty(p2)) return 0;
	
	int i;
	
	for(i=0;i< p2.size();++i)
	{
		if( ! ( p1[i] < p2[i] ) ) return 0;
	}
	return 1;
	
};

	
/*trigonometry*/

inline interval iSin(interval i)
{
	interval j=i%DEUX_PI;
	REEL s1,s2,b1,b2;
	if (i.length()>DEUX_PI) return interval(-1,1);

	s1=sin(j.getFirst());   s2=sin(j.getLast());
	if ((j.getFirst() < TROIS_PI2)&&(j.getLast() > TROIS_PI2)) b1=-1.0;
	else if ((j.getFirst() < SEPT_PI2)&&(j.getLast() > SEPT_PI2)) b1=-1.0;
	else b1=min(s1,s2);

	if ((j.getFirst() < PI2)&&(j.getLast() > PI2)) b2=1.0;
	else if ((j.getFirst() < CINQ_PI2)&&(j.getLast() > CINQ_PI2)) b2=1.0;
	else b2=max(s1,s2);
	return interval(b1,b2);
	
}
inline interval iCos(interval i)
{
	return iSin(i+PI2);
};
inline interval iAtan(interval i)
{
	return interval(atan(i.getFirst()),atan(i.getLast()));
};

///////////////////////////////////////////////////////////////////////////////////////////
//////////////WARNING : following function are written for the ////////////////////////////
//////////////dielectric spectroscopy data fit only, and assume////////////////////////////
////////////// that every angle is beetwen 0 and PI/2...       ////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////

inline interval iSinInv(interval i)
{
	return interval(asin(i.first), asin(i.last));
};

inline interval iCosInv(interval i)
{
	return interval(acos(i.last),acos(i.first));
};

inline interval iLog(interval i)
{
	return (interval(log(i.first), log(i.last)));
};

inline interval iTan(interval i)
{
	return interval(tan(i.first),tan(i.last));
};





	


#endif
